#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#define DO_SYNTH

/* McBSP Register Offsets. */
#define MCBSP_DRR_OFFSET            0x00
#define MCBSP_DXR_OFFSET            0x08
#define MCBSP_SPCR2_OFFSET          0x10
#define MCBSP_SPCR1_OFFSET          0x14
#define MCBSP_RCR2_OFFSET           0x18
#define MCBSP_RCR1_OFFSET           0x1C
#define MCBSP_XCR2_OFFSET           0x20
#define MCBSP_XCR1_OFFSET           0x24
#define MCBSP_SRGR2_OFFSET          0x28
#define MCBSP_SRGR1_OFFSET          0x2C
#define MCBSP_MCR2_OFFSET           0x30
#define MCBSP_MCR1_OFFSET           0x34
#define MCBSP_RCERA_OFFSET          0x38
#define MCBSP_RCERB_OFFSET          0x3C
#define MCBSP_XCERA_OFFSET          0x40
#define MCBSP_XCERB_OFFSET          0x44
#define MCBSP_PCR0_OFFSET           0x48

#define MCBSP_REV_OFFSET            0x7C
#define MCBSP_RINTCLR_OFFSET        0x80
#define MCBSP_XINTCLR_OFFSET        0x84
#define MCBSP_ROVFLCLR_OFFSET       0x88
#define MCBSP_SYSCONFIG_OFFSET      0x8C
#define MCBSP_THRSH2_OFFSET         0x90
#define MCBSP_THRSH1_OFFSET         0x94
#define MCBSP_IRQSTATUS_OFFSET      0xA0

#define MCBSP_IRQSTATUS_OFFSET      0xA0
#define MCBSP_IRQENABLE_OFFSET      0xA4
#define MCBSP_WAKEUPEN_OFFSET       0xA8
#define MCBSP_XCCR_OFFSET           0xAC
#define MCBSP_RCCR_OFFSET           0xB0
#define MCBSP_XBUFFSTAT_OFFSET      0xB4
#define MCBSP_RBUFFSTAT_OFFSET      0xB8

#define BITSPERSAMPLE   16

typedef unsigned int U32;
#define out_regl(a,d) *((volatile unsigned int *)(a)) = d


/* Reset the McBSP. */
static void reset_mcbsp (U32 mcbsp_base_addr, U32 enable)
{
    /*
    ** Checking whether McBSP is providing the clock (Word and bit CLK)
    ** or McBSP is master.
    */
    if (((*(volatile U32 *)(mcbsp_base_addr + MCBSP_PCR0_OFFSET)) & 0x0F00)
        != 0x0F00)
    {
        /*
        ** If McBSP is not master, then TX/RX/SCG are not enabled
        ** and disabled between each call to the macbsp_write/read functions.
        */
        return;
    }
    else
    {
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_DXR_OFFSET))= 0x0;
        if (enable)
        {
            /* The serial port transmitter is enabled. */
            *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR2_OFFSET)) |= 0x0001;

            /* The serial port receiver is enabled. */
            *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR1_OFFSET)) |= 0x0001;

            /*
            ** Sample rate generator is pulled out of reset.
            ** Frame counters are loaded with their programmed values.
            */
            *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR2_OFFSET)) |= 0x00C0;
        }
        else
        {
            /*
            ** The serial port transmitter is disabled and in reset state.
            ** Frame-synchronization logic is reset.
            */
            *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR2_OFFSET))
                &= ~(0x00C1);

            /* The serial port receiver is disabled and in reset state. */
            *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR1_OFFSET))
                &= ~(0x0001);
        }
    }
}


/* Writes TX data register by polling method. */
static int write_data_mcbsp(U32 mcbsp_base_addr, U32 data)
{
    U32 ret_status = 0;
    U32 attempts;

    *((volatile U32 *)(mcbsp_base_addr + MCBSP_DXR_OFFSET)) = data;

    /* If frame sync error - clear the error. */
    if (*((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR2_OFFSET)) & 0x08)
    {
        printf("Frame sync error.\n");
        /* Clear error. */
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR2_OFFSET)) &= ~(0x08);
    }

    /* Wait for transmit confirmation. */
    attempts = 0;
    while  ((*((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR2_OFFSET)) & 0x2)
            != 0x2)
    {
        if (attempts++ > 1000)
        {
            /* The attempt failed, so reset the Tx. */
            *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR2_OFFSET)) &= ~(0x1);

            /* Enable Tx. */
            *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR2_OFFSET)) |= 0x1;

            printf("McBSP write failed.\n");
            return -1;
        }
    }

    return ret_status;
}

int main(int argc, char *argv[])
{
	volatile unsigned int *regs;
	unsigned int mcbsp_base_addr, sample;
	int ret, memdev;
#ifdef DO_SYNTH
	int count = 0;
#else
	FILE *f;

	f = fopen(argv[1], "rb");
	if (f == NULL)
	{
		printf("no file.\n");
		return 0;
	}
#endif

	memdev = open("/dev/mem", O_RDWR|O_SYNC);
	if (memdev == -1)
	{
		printf("no memdev\n");
		return 1;
	}

	// 0x480022D8 ~ CONTROL_DEVCONF1
	// 0x49024000 ~ McBSP3
  	regs = mmap(0, 0x10000, PROT_WRITE|PROT_READ, MAP_SHARED, memdev, 0x48000000);
	if (regs == MAP_FAILED)
	{
		printf("mmap failed\n");
		return 1;
	}

  	mcbsp_base_addr = (unsigned int)mmap(0, 0x2000, PROT_WRITE|PROT_READ, MAP_SHARED, memdev, 0x49024000);
	if ((int)mcbsp_base_addr == (int)MAP_FAILED)
	{
		printf("mmap failed\n");
		return 1;
	}

	/* ----- clocks ----- */

	/*
	 ** Configure the CLKS input to 96M functional clk from PRCM
	 ** and not MCBSP.CLKS pin.(this is for McBSP2)
	 ** CONTROL_DEVCONF0 ,set bit 6 to zero.
	 */
	regs[0x2274>>2] &= 0xFFFFFFBF;

	// CONTROL_DEVCONF1 For McBSP3 we want to use the MCBSP.CLKS pin
	regs[0x22D8>>2] |= 1;

	/* CM_CLKSEL1_PLL 96 Mhz selected. */
	regs[0x4d40>>2] &= 0xFFFFFFF7;

	/* CM_FCLKEN_PER Enable the functional clock of McBSP3. */
	regs[0x5000>>2] |= 0x00000002;

	/* CM_ICLKEN_PER Enable the Interface clock of McBSP3 */
	regs[0x5010>>2] |= 0x00000002;

	/* McBSP3 selected. */
	regs[0x5030>>2] |= 0x00000002;

	/* ----- McBSP ----- */

	/* Make all registers to zero to start with. */
	out_regl((mcbsp_base_addr + MCBSP_SPCR2_OFFSET), 0);
	out_regl((mcbsp_base_addr + MCBSP_SPCR1_OFFSET), 0);
//	out_regl((mcbsp_base_addr + MCBSP_PCR0_OFFSET), 0);
	out_regl((mcbsp_base_addr + MCBSP_XCR1_OFFSET), 0);
	out_regl((mcbsp_base_addr + MCBSP_RCR1_OFFSET), 0);
	out_regl((mcbsp_base_addr + MCBSP_XCR2_OFFSET), 0);
	out_regl((mcbsp_base_addr + MCBSP_RCR2_OFFSET), 0);
	out_regl((mcbsp_base_addr + MCBSP_SRGR1_OFFSET), 0);
	out_regl((mcbsp_base_addr + MCBSP_SRGR2_OFFSET), 0);

	/* Disable TX and RX DMA. */
	out_regl((mcbsp_base_addr + MCBSP_XCCR_OFFSET), 0);
	out_regl((mcbsp_base_addr + MCBSP_RCCR_OFFSET), 0);
	out_regl((mcbsp_base_addr + MCBSP_THRSH2_OFFSET), 0);
	out_regl((mcbsp_base_addr + MCBSP_THRSH1_OFFSET), 0);
	out_regl((mcbsp_base_addr + MCBSP_IRQENABLE_OFFSET), 0);
	out_regl((mcbsp_base_addr + MCBSP_IRQSTATUS_OFFSET), 0x7fbf); // clear IRQs

        /* Set the audio configuration. */

        /*
        ** Codec is slave McBSP is master
        ** Enable the following bits FSXM, FSRM, CLKXM, CLKRM, CLKXP, CLKRP.
        */
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_PCR0_OFFSET)) = 0x0f0C;//0x0f00; //0x0f03;
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR1_OFFSET)) |=  0x0008;
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR2_OFFSET)) |=  0x0008;

        /* Tx word len1= 16bits, 1 word per frame. */
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_XCR1_OFFSET))  |=  0x40;

        /*
        ** Dual phase frame
        ** Tx word len2 = 16 bits, 1 word per frame.
        ** No companding data starts with MSB first
        ** Transmit frame-synchronization pulses after the first are ignored.
        ** Tx Data delay - 1bit.
        */
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_XCR2_OFFSET))  |=  0x8041;

        /* Rx word len1 = 16bits, 1 word per frame. */
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_RCR1_OFFSET))  |=  0x40;

        /*
        ** Dual phase frame
        ** Rx word len2=16 bits ,1 word per frame.
        ** No companding data starts with MSB first
        ** Rx Data delay- 1bit.
        */
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_RCR2_OFFSET))  |=  0x8041;

        /* Configure the sample generator. */

        /*
        ** McBSP sample rate generator register 1
        ** Frame pulse width is 16 clk periods
        ** Sample rate generator clock divider - 70
        ** 44.1Khz
	*/
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_SRGR1_OFFSET))
            = ((BITSPERSAMPLE - 1) << 8) | (8 - 1); // BCLK = 32fs

        /* Next frame-sync signal becomes active after 32 clock periods. */
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_SRGR2_OFFSET)) =
//		0x2000 |
		0x1000 | ((BITSPERSAMPLE * 2) - 1);

        /* Enable the sample rate generator. */
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR2_OFFSET)) |= 0x00C0;

        /* Enable the Tx, Rx. */
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR1_OFFSET)) |= 1;
        *((volatile U32 *)(mcbsp_base_addr + MCBSP_SPCR2_OFFSET)) |= 1;

        usleep(1000);

	/* Reset the device. */
	reset_mcbsp(mcbsp_base_addr, 1);

	/* must restore ?DMAEN bits to defaults */
	out_regl((mcbsp_base_addr + MCBSP_XCCR_OFFSET), 8);
	out_regl((mcbsp_base_addr + MCBSP_RCCR_OFFSET), 8);

        usleep(1000);

	/* write */
	sample = 0;
	while (1)
	{
#ifndef DO_SYNTH
		ret = fread(&sample, 1, 2, f);
		if (ret != 2) break;
#else
		sample = (count & 0x80) ? 65536-10000 : 10000;
		count++;
		if (count > 200*1024) break;
#endif

		ret = write_data_mcbsp(mcbsp_base_addr, sample);
		if (ret != 0) break;
	}

	/* Reset the device. */
	reset_mcbsp(mcbsp_base_addr, 0);

	munmap((void *)regs, 0x10000);
	munmap((void *)mcbsp_base_addr, 0x2000);
	close(memdev);
#ifndef DO_SYNTH
	fclose(f);
#endif

	return 0;
}

