#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>

static u32 test_addr;

#define CTL_TEX_REMAP (1 << 28)

// struct mm_struct *mm = vma->vm_mm;
// pgd_t *pgd = pgd_offset(mm, addr);

// input: v = TEX[4:2] | C[1] | B[0]
// S is shareable bit from TLB, can be remapped
// on Cortex-A8 it just disables cache (what about buffer?)
static char *parse_tex_cb(char *p, u32 v, int s)
{
	static const char *cache_attrs4[4] = { "non-cacheable", "WB-WA", "WT-noWA", "WB-noWA" };
	u32 prrr, nmrr = 0;
	u32 control = 0;
	int type;

	asm ("mrc p15, 0, %0, c1, c0, 0" : "=r"(control));
	// remap registers
	asm ("mrc p15, 0, %0, c10, c2, 0" : "=r"(prrr)); // primary region RR
	asm ("mrc p15, 0, %0, c10, c2, 1" : "=r"(nmrr)); // normal memory RR

	if (control & CTL_TEX_REMAP) {
		// S (shareable) bit remapping
		char s_normal[2] = { (prrr >> 18) & 1, (prrr >> 19) & 1 };
		char s_device[2] = { (prrr >> 16) & 1, (prrr >> 17) & 1 };

		v &= 7;
		type = (prrr >> v * 2) & 3;
		switch (type) {
		case 0:
			p += sprintf(p, "strongly-ordered ");
			break;
		case 1:
			p += sprintf(p, "device ");
			s = s_device[s];
			break;
		case 3:
			p += sprintf(p, "reserved/normal ");
		case 2:
			s = s_normal[s];
			p += sprintf(p, "inner-%s-outer-%s ",
				cache_attrs4[(nmrr >> v * 2) & 3],
				cache_attrs4[(nmrr >> (v * 2 + 16)) & 3]);
		}
	}
	else if (v & 0x10) { // TEX[2] set
		p += sprintf(p, "inner-%s-outer-%s ",
			cache_attrs4[v & 3], cache_attrs4[(v >> 2) & 3]);
	}
	else {
		switch (v) {
		case 0x00: p += sprintf(p, "strongly-ordered "); s = 1; break;
		case 0x01: p += sprintf(p, "shareable-device "); s = 1; break;
		case 0x02: p += sprintf(p, "inner-outer-WT-noWA "); break;
		case 0x03: p += sprintf(p, "inner-outer-WB-noWA "); break;
		case 0x04: p += sprintf(p, "inner-outer-non-cacheable "); break;
		case 0x06: p += sprintf(p, "implementation-defined "); break;
		case 0x07: p += sprintf(p, "inner-outer-WB-WA "); break;
		case 0x08: p += sprintf(p, "non-shareable-device "); s = 0; break;
		default:   p += sprintf(p, "reserved "); break;
		}
	}

	if (s)
		p += sprintf(p, "shareable");

	return p + sprintf(p, "\n");
}

static int proc_read(char *page, char **start, off_t off, int count,
		int *eof, void *data)
{
	int verbose = test_addr & 1;
	u32 ttb0 = 0, ttb1 = 0, ttbc = 0;
	u32 desc1, desc2;
	u32 *pgtable, *cpt;
	u32 tex_cb = 0;
	u32 phys;
	char *p = page;
	int len;

	asm ("mrc p15, 0, %0, c2, c0, 0" : "=r"(ttb0)); // TLB base
	asm ("mrc p15, 0, %0, c2, c0, 1" : "=r"(ttb1));
	asm ("mrc p15, 0, %0, c2, c0, 2" : "=r"(ttbc)); // TLB control

	p += sprintf(p, "%08x: ", test_addr);
	if (verbose) {
		p += sprintf(p, "\n");
		p += sprintf(p, "TLB: %08x %08x %08x\n", ttb0, ttb1, ttbc);
	}

	pgtable = __va(ttb0 & 0xffffc000);
	desc1 = pgtable[test_addr >> 20];

	if (verbose)
		p += sprintf(p, "l1: %08x ", desc1);
	switch (desc1 & 3) {
	case 0:
		p += sprintf(p, "l1_fault\n");
		goto out;
	case 1:
		if (verbose)
			p += sprintf(p, "page\n");
		break;
	case 2:
		if (verbose)
			p += sprintf(p, "%ssection\n", desc1 & (1<<18) ? "super" : "");
		if (desc1 & (1<<18))
			phys = (desc1 & 0xff000000) | (test_addr & 0x00ffffff);
		else
			phys = (desc1 & 0xfff00000) | (test_addr & 0x000fffff);
		p += sprintf(p, "%08x ", phys);
		tex_cb = ((desc1 >> 2) & 0x03) | ((desc1 >> 10) & 0x1c);
		p = parse_tex_cb(p, tex_cb, (desc1 >> 16) & 1);
		goto out;
	case 3:
		p += sprintf(p, "reserved\n");
		goto out;
	}

	cpt = __va(desc1 & 0xfffffc00);
	desc2 = cpt[(test_addr >> 12) & 0xff];

	// level2
	if (verbose)
		p += sprintf(p, "l2: %08x ", desc2);
	switch (desc2 & 3) {
	case 0:
		p += sprintf(p, "l2_fault\n");
		break;
	case 1:
		if (verbose)
			p += sprintf(p, "large_page\n");
		phys = (desc2 & 0xffff0000) | (test_addr & 0x0000ffff);
		p += sprintf(p, "%08x ", phys);
		tex_cb = ((desc2 >> 2) & 0x03) | ((desc2 >> 10) & 0x1c);
		p = parse_tex_cb(p, tex_cb, (desc2 >> 10) & 1);
		break;
	case 2:
	case 3:
		if (verbose)
			p += sprintf(p, "small_page\n");
		phys = (desc2 & 0xfffff000) | (test_addr & 0x00000fff);
		p += sprintf(p, "%08x ", phys);
		tex_cb = ((desc2 >> 2) & 0x03) | ((desc2 >> 4) & 0x1c);
		p = parse_tex_cb(p, tex_cb, (desc2 >> 10) & 1);
		break;
	}

out:
	len = (p - page) - off;
	if (len < 0)
		len = 0;

	*eof = (len <= count) ? 1 : 0;
	*start = page + off;

	return len;
}

static int proc_write(struct file *file, const char __user *buffer,
		unsigned long count, void *data)
{
	char buff[32];
	unsigned long val;
	int ret;

	count = strncpy_from_user(buff, buffer,
			count < sizeof(buff) ? count : sizeof(buff) - 1);
	buff[count] = 0;

	ret = strict_strtoul(buff, 0, &val);
	if (ret < 0) {
		printk(KERN_ERR "error %i parsing %s\n", ret, buff);
		return ret;
	}

	test_addr = val;

	return count;
}

static int init_my_module(void)
{
	struct proc_dir_entry *pret;
	u32 control = 0, l2_aux = 0;
	u32 prrr = 0, nmrr = 0;

	printk(KERN_INFO "tlbtest loaded.\n");

	asm ("mrc p15, 0, %0, c1, c0, 0" : "=r"(control));
	asm ("mrc p15, 1, %0, c9, c0, 2" : "=r"(l2_aux));
	asm ("mrc p15, 0, %0, c10, c2, 0" : "=r"(prrr)); // primary region RR
	asm ("mrc p15, 0, %0, c10, c2, 1" : "=r"(nmrr)); // normal memory RR
	printk(KERN_INFO "TEX remap is %s.\n",
		(control & CTL_TEX_REMAP) ? "enabled" : "disabled");
	printk(KERN_INFO "L2 cache uses %s attributes.\n",
		(l2_aux & (1 << 16)) ? "inner" : "outer");
	// 000a81a8, 40e040e0
	printk(KERN_INFO "PRRR: %08x, NMRR: %08x\n", prrr, nmrr);

	pret = create_proc_entry("tlbtest",
			S_IWUSR | S_IRUGO, NULL);
	if (pret == NULL) {
		printk(KERN_ERR "can't create proc\n");
		return -1;
	}

	pret->read_proc = proc_read;
	pret->write_proc = proc_write;

	return 0;
}


static void cleanup_my_module(void)
{
	remove_proc_entry("tlbtest", NULL);
	printk("tlbtest unloaded.\n");
}


module_init(init_my_module);
module_exit(cleanup_my_module);

MODULE_AUTHOR("notaz");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("TLB test");

